@@ -11,10 +11,9 @@ from jurigged import CodeFile, registry
 from jurigged.recode import virtual_file
 
 from .feat.edit import edit
-from .session import current_session, new_evalid, threads
+from .session import current_session, new_evalid, threads, SnektalkInterrupt
 from .version import version
 
-
 cmd_rx = re.compile(r"/([^ \n]+)([ \n].*)?", re.MULTILINE | re.DOTALL)
 
 
@@ -129,6 +128,7 @@ class Evaluator:
     @safe_fail
     def command_edit(self, expr, glb, lcl):
         expr = expr.lstrip()
+        self.session.queue(command="echo", value=f"/edit {expr}", process=False)
         obj = self.eval(expr, glb, lcl)
         self.session.queue_result(
             edit(obj, evaluator=self, autofocus=True), type="expression"
@@ -196,52 +196,6 @@ class Evaluator:
 
         threads.run_in_thread(run, session=current_session())
 
-    def command_attach(self, expr, glb, lcl):
-        tname = expr.strip()
-        self.session.queue(
-            command="echo", value=f"/attach {tname}", process=False
-        )
-        if tname == "main":
-            current_session().push_owner(None)
-        elif tname in threads.threads:
-            thread = threads.threads[tname]
-            current_session().push_owner(thread)
-        else:
-            self.session.queue(
-                command="result",
-                value=f"No thread named {tname}"
-                if tname
-                else "Please provide the name of the thread to attach to",
-                type="exception",
-            )
-
-    def command_detach(self, expr, glb, lcl):
-        self.session.queue(command="echo", value=f"/detach", process=False)
-        assert not expr.strip()
-        current_session().pop_owner()
-
-    def command_kill(self, expr, glb, lcl):
-        tname = expr.strip()
-        self.session.queue(
-            command="echo", value=f"/kill {tname}", process=False
-        )
-        if tname in threads.threads:
-            thread = threads.threads[tname]
-            thread.kill()
-            self.session.queue(
-                command="result",
-                value=f"Sent an exception to {tname}. It should terminate as soon as possible.",
-                type="print",
-            )
-        else:
-            self.session.queue(
-                command="result",
-                value=f"No thread named {tname}"
-                if tname
-                else "Please provide the name of the thread to kill",
-                type="exception",
-            )
-
     def command_quit(self, expr, glb, lcl):
         self.session.queue_result(H.div("/quit"), type="echo")
         self.session.queue_result(
@@ -293,14 +247,20 @@ class Evaluator:
         )
         try:
             while True:
-                prompt = H.span["snek-input-mode-python"](self.prompt)
-                with self.session.prompt(prompt) as cmd:
-                    if cmd["command"] == "expr":
-                        expr = cmd["expr"]
-                        if expr.strip():
-                            self.dispatch(expr)
-                    elif cmd["command"] == "noop":
-                        pass
+                try:
+                    prompt = H.span["snek-input-mode-python"](self.prompt)
+                    with self.session.prompt(prompt) as cmd:
+                        if cmd["command"] == "expr":
+                            expr = cmd["expr"]
+                            if expr.strip():
+                                try:
+                                    self.dispatch(expr)
+                                except SnektalkInterrupt as exc:
+                                    print(exc)
+                        elif cmd["command"] == "noop":
+                            pass
+                except SnektalkInterrupt:
+                    continue
         except StopEvaluator:
             return
 
