@@ -4,6 +4,7 @@ import types
 from types import FunctionType, MethodType, ModuleType, SimpleNamespace
 
 from hrepr import H, Hrepr, Tag, hrepr, standard_html
+from hrepr.std import _extract_as
 
 from .debug import SnekTalkDb
 from .feat.edit import edit
@@ -26,10 +27,18 @@ class PrintSequence(tuple):
         )
 
 
-def snekprint(*args, toplevel=False, **kwargs):
+def snekprint(*args, toplevel=False, to=None, **kwargs):
     sess = current_print_session()
     if sess is None:
         orig_print(*args)
+    elif to is not None:
+        if toplevel:
+            html = hrepr(*args, **kwargs)
+        else:
+            html = hrepr(PrintSequence(args), **kwargs)
+        sess.queue(
+            command="insert", value=html, target=to,
+        )
     else:
         if all(isinstance(arg, str) for arg in args):
             html = H.div["snek-print-str"](" ".join(args))
@@ -52,6 +61,23 @@ def snekprint_override(*args, toplevel=False, file=None, std=False, **kwargs):
     builtins.print = snekprint_override
 
 
+def insert_at(target, value, index=None, **kwargs):
+    html = hrepr(value, **kwargs)
+    sess = current_session()
+    sess.queue(command="insert", value=html, target=target, index=index)
+
+
+def fill_at(target, value, **kwargs):
+    html = hrepr(value, **kwargs)
+    sess = current_session()
+    sess.queue(command="fill", value=html, target=target)
+
+
+def clear_at(target):
+    sess = current_session()
+    sess.queue(command="clear", target=target)
+
+
 def help_placeholder(*args, **kwargs):
     print(
         H.span("The help command is not available in snektalk for the moment.")
@@ -322,13 +348,40 @@ class SnekTalkHrepr(Hrepr):
 #####################################
 
 
-sktk = SimpleNamespace(edit=edit, help=None, imp=Importer(), mod=mod,)
+sktk = SimpleNamespace(
+    edit=edit,
+    help=None,
+    imp=Importer(),
+    mod=mod,
+    fill_at=fill_at,
+    insert_at=insert_at,
+    clear_at=clear_at,
+)
 
 
 def snekbreakpoint():
     SnekTalkDb().set_trace()
 
 
+@standard_html.variant(
+    initial_state={"hjson": sktk_hjson, "requirejs_resources": []}
+)
+def sktk_html(self, node: type(H.include)):
+    _, children, data = _extract_as(self, node, "include", path=None, type=None)
+    if data.type is None or data.path is None:
+        raise TypeError("H.include must have a type and a path")
+
+    path = os.path.expanduser(data.path)
+    path = os.path.abspath(path)
+
+    if data.type == "text/css":
+        return H.link(rel="stylesheet", href=f"/fs{path}")
+    elif data.type == "text/javascript":
+        return H.script(type="text/javascript", src=f"/fs{path}")
+    else:
+        raise TypeError(f"Cannot include type '{data.type}'")
+
+
 def inject():
     builtins.help = help_placeholder
     builtins.print = snekprint_override
@@ -337,9 +390,5 @@ def inject():
     builtins.print0 = orig_print
     builtins.sktk = sktk
     hrepr.configure(
-        mixins=SnekTalkHrepr,
-        postprocess=wrap_onclick,
-        backend=standard_html.copy(
-            initial_state={"hjson": sktk_hjson, "requirejs_resources": []}
-        ),
+        mixins=SnekTalkHrepr, postprocess=wrap_onclick, backend=sktk_html,
     )
